from smartrent.api import async_login
from smartrent.lock import DoorLock
from smartrent.thermostat import Thermostat
from smartrent.sensor import LeakSensor
from smartrent.switch import BinarySwitch
from smartrent.utils import Client

__all__ = [
    "async_login",
    "DoorLock",
    "Thermostat",
    "LeakSensor",
    "BinarySwitch",
    "Client",
]



import logging
from typing import List, Union

import aiohttp

from .lock import DoorLock
from .thermostat import Thermostat
from .switch import BinarySwitch
from .sensor import LeakSensor
from .utils import Client

_LOGGER = logging.getLogger(__name__)


class API:
    """
    Represents overall SmartRent api

    ``email`` is the email address for your SmartRent account

    ``password`` you know what it is

    ``aiohttp_session`` (optional) uses the aiohttp_session that is passed in
    """

    def __init__(
        self, email: str, password: str, aiohttp_session: aiohttp.ClientSession = None
    ):
        self._device_list = []
        self._email = email
        self._password = password
        self._session = aiohttp_session

        self.client: Client = Client(email, password, aiohttp_session)

    async def async_fetch_devices(self):
        """
        Fetches list of devices by calling SmartRent api
        """
        _LOGGER.info("Fetching devices via API...")
        await self.client.async_refresh_token()
        data = await self.client.async_get_devices_data()
        _LOGGER.info("Got devices!")

        for device in data:
            device_id = device.get("id")
            device_type = device.get("type")

            device_object: Union[Thermostat, DoorLock, BinarySwitch, LeakSensor] = None

            if device_type == "thermostat":
                device_object = Thermostat(device_id, self.client)

            elif device_type == "entry_control":
                device_object = DoorLock(device_id, self.client)

            elif device_type == "switch_binary":
                device_object = BinarySwitch(device_id, self.client)

            elif device_type == "sensor_notification":
                attr_names = [attr.get("name") for attr in device.get("attributes")]

                if "leak" in attr_names:
                    device_object = LeakSensor(device_id, self.client)

            if device_object:
                # pass in intial device config
                device_object._fetch_state_helper(device)

                # add device to device_list
                self._device_list.append(device_object)

    def get_device_list(
        self,
    ) -> List[Union[DoorLock, Thermostat, BinarySwitch, LeakSensor]]:
        """
        Gets list of all devices found
        """
        return self._device_list

    def get_locks(self) -> List[DoorLock]:
        """
        Gets list of DoorLocks
        """
        return [x for x in self._device_list if isinstance(x, DoorLock)]

    def get_thermostats(self) -> List[Thermostat]:
        """
        Gets list of Thermostats
        """
        return [x for x in self._device_list if isinstance(x, Thermostat)]

    def get_switches(self) -> List[BinarySwitch]:
        """
        Gets list of BinarySwitches
        """
        return [x for x in self._device_list if isinstance(x, BinarySwitch)]

    def get_leak_sensors(self) -> List[LeakSensor]:
        """
        Gets list of LeakSensors
        """
        return [x for x in self._device_list if isinstance(x, LeakSensor)]


async def async_login(
    email: str, password: str, aiohttp_session: aiohttp.ClientSession = None
) -> API:
    """
    Logs into SmartRent and retruns an ``API`` object.

    Prepopulates ``API`` object with devices.

    ``email`` is the email address for your SmartRent account

    ``password`` you know what it is

    ``aiohttp_session`` (optional) uses the aiohttp_session that is passed in
    """

    smart_rent_api = API(email, password, aiohttp_session)
    await smart_rent_api.async_fetch_devices()

    return smart_rent_api



from typing import Optional
import logging

from .device import Device
from .utils import Client

_LOGGER = logging.getLogger(__name__)


class Thermostat(Device):
    """
    Represents Thermostat SmartRent device
    """

    def __init__(self, device_id: int, client: Client):
        super().__init__(device_id, client)
        self._mode = None
        self._fan_mode = None
        self._cooling_setpoint = None
        self._heating_setpoint = None
        self._current_humidity = None
        self._current_temp = None

    def get_mode(self) -> Optional[str]:
        """
        Gets mode from thermostat
        """
        return self._mode

    def get_fan_mode(self) -> Optional[str]:
        """
        Gets fan mode from thermostat
        """
        return self._fan_mode

    def get_cooling_setpoint(self) -> Optional[int]:
        """
        Gets cooling setpoint from thermostat
        """
        return self._cooling_setpoint

    def get_heating_setpoint(self) -> Optional[int]:
        """
        Gets heating setpoint from thermostat
        """
        return self._heating_setpoint

    def get_current_humidity(self) -> Optional[int]:
        """
        Gets current humidity from thermostat
        """
        return self._current_humidity

    def get_current_temp(self) -> Optional[int]:
        """
        Gets current temperature from thermostat
        """
        return self._current_temp

    async def async_set_heating_setpoint(self, value: int):
        """
        Sets heating setpoint

        ``value`` str or int representing temperature to set
        """
        await self._client._async_send_command(
            self, attribute_name="heating_setpoint", value=value
        )

        self._heating_setpoint = int(value)

    async def async_set_cooling_setpoint(self, value: int):
        """
        Sets cooling setpoint

        ``value`` str or int representing temperature to set
        """
        await self._client._async_send_command(
            self, attribute_name="cooling_setpoint", value=value
        )

        self._cooling_setpoint = int(value)

    async def async_set_mode(self, mode: str):
        """
        Sets thermostat mode

        ``mode`` str. One of ``['aux_heat', 'heat', 'cool', 'auto', 'off']``
        """
        if mode not in ["aux_heat", "heat", "cool", "auto", "off"]:
            return

        await self._client._async_send_command(self, attribute_name="mode", value=mode)

        self._mode = mode

    async def async_set_fan_mode(self, fan_mode: str):
        """
        Sets thermostat fan mode

        ``value`` str. One of ``['auto', 'on']``
        """
        if fan_mode not in ["on", "auto"]:
            return

        await self._client._async_send_command(
            self,
            attribute_name="fan_mode",
            value=fan_mode,
        )

        self._fan_mode = fan_mode

    def _fetch_state_helper(self, data: dict):
        """
        Called when ``_async_fetch_state`` returns info

        ``data`` is dict of info passed in by ``_async_fetch_state``
        """
        self._name = data["name"]

        attrs = self._structure_attrs(data["attributes"])

        def float_to_int(x: Optional[str]):
            if x and x != "None":
                return int(float(x))

        self._current_temp = float_to_int(attrs.get("current_temp"))

        self._cooling_setpoint = float_to_int(attrs.get("cooling_setpoint"))
        self._heating_setpoint = float_to_int(attrs.get("heating_setpoint"))

        self._current_humidity = float_to_int(attrs.get("current_humidity"))

        self._mode = attrs["mode"]
        self._fan_mode = attrs.get("fan_mode")

    def _update_parser(self, event: dict) -> None:
        """
        Called when ``_async_update_state`` returns info

        ``event`` dict passed in from ``_async_update_state``
        """
        _LOGGER.info("Updating Thermostat")
        if event.get("name") == "current_humidity":
            self._current_humidity = int(event.get("last_read_state"))

        if event.get("name") == "current_temp":
            self._current_temp = int(event.get("last_read_state"))

        if event.get("name") == "heating_setpoint":
            self._heating_setpoint = int(event.get("last_read_state"))

        if event.get("name") == "cooling_setpoint":
            self._cooling_setpoint = int(event.get("last_read_state"))

        if event.get("name") == "mode":
            self._mode = event.get("last_read_state")

        if event.get("name") == "fan_mode":
            self._fan_mode = event.get("last_read_state")



from typing import Optional
import logging

from .device import Device
from .utils import Client

_LOGGER = logging.getLogger(__name__)


class BinarySwitch(Device):
    """
    Represents BinarySwitch SmartRent device
    """

    def __init__(self, device_id: int, client: Client):
        super().__init__(device_id, client)
        self._on = None

    def get_on(self) -> Optional[bool]:
        """
        Gets state from switch
        """
        return self._on

    async def async_set_on(self, value: bool):
        """
        Sets state for switch
        """
        self._on = value

        # Convert to lowercase just like SmartRent website does
        value = str(value).lower()

        await self._client._async_send_command(self, attribute_name="on", value=value)

    def _fetch_state_helper(self, data: dict):
        """
        Called when ``_async_fetch_state`` returns info

        ``data`` is dict of info passed in by ``_async_fetch_state``
        """
        self._name = data["name"]

        attrs = self._structure_attrs(data["attributes"])

        self._on = bool(attrs["on"] == "true")

    def _update_parser(self, event: dict):
        """
        Called when ``_async_update_state`` returns info

        ``event`` dict passed in from ``_async_update_state``
        """
        _LOGGER.info("Updating Switch")

        if event.get("name") == "on":
            self._on = bool(event["last_read_state"] == "true")



import asyncio
import logging
import json
import traceback
from typing import List, Set, TYPE_CHECKING

import aiohttp
import websockets

if TYPE_CHECKING:
    from smartrent.device import Device

_LOGGER = logging.getLogger(__name__)

SMARTRENT_BASE_URI = "https://control.smartrent.com/api/v2/"
SMARTRENT_SESSIONS_URI = SMARTRENT_BASE_URI + "sessions"
SMARTRENT_TOKENS_URI = SMARTRENT_BASE_URI + "tokens"
SMARTRENT_HUBS_URI = SMARTRENT_BASE_URI + "hubs"
SMARTRENT_HUBS_ID_URI = SMARTRENT_BASE_URI + "hubs/{}/devices"

SMARTRENT_WEBSOCKET_URI = (
    "wss://control.smartrent.com/socket/websocket?token={}&vsn=2.0.0"
)
JOINER_PAYLOAD = '["null", "null", "devices:{device_id}", "phx_join", {{}}]'
COMMAND_PAYLOAD = (
    '["null", "null", "devices:{device_id}", "update_attributes", '
    '{{"device_id": {device_id}, '
    '"attributes": [{{"name": "{attribute_name}", "value": "{value}"}}]}}]'
)


class SmartRentError(Exception):
    """
    Base error for SmartRent
    """


class InvalidAuthError(SmartRentError):
    """
    Error related to invalid auth
    """


class Client:
    """
    Represents Client for SmartRent http api.
    Usually shared between multiple devices for best performance.

    ``email`` is the email address for your SmartRent account

    ``password`` you know what it is

    ``aiohttp_session`` (optional) uses the aiohttp_session that is passed in
    """

    def __init__(
        self,
        email: str,
        password: str,
        aiohttp_session: aiohttp.ClientSession = None,
    ):
        self._email = email
        self._password = password

        self._im_session_owner = not bool(aiohttp_session)
        self._aiohttp_session = (
            aiohttp_session if aiohttp_session else aiohttp.ClientSession()
        )
        self._token = None
        self._refresh_token = None
        self._token_exp_time = None

        self._subscribed_devices: Set["Device"] = set()
        self._updater_task = None
        self._ws = None

    def __del__(self):
        """
        Handles delete of aiohttp session if class is tasked with it
        """
        if not self._aiohttp_session.closed and self._im_session_owner:
            _LOGGER.debug(
                "%s: closing aiohttp session %s", str(self), self._aiohttp_session
            )

            try:
                _LOGGER.debug("Finding running event loop")
                current_loop = asyncio.get_running_loop()
                current_loop.create_task(self._aiohttp_session.close())

            except RuntimeError:
                _LOGGER.debug("Making new event loop")
                new_loop = asyncio.new_event_loop()
                new_loop.run_until_complete(self._aiohttp_session.close())

        if self._updater_task:
            _LOGGER.info("Stopping updater task")
            self._updater_task.cancel()
            self._ws = None

    async def async_get_devices_data(self) -> List[dict]:
        """
        Gets list of device dictonaries from SmartRent's api.
        Also handles retry if token is bad
        """
        if not self._token:
            await self.async_refresh_token()

        try:
            res = await self._async_get_devices_data()
        except InvalidAuthError:
            _LOGGER.warning("InvalidAuth detected. Trying again with updated token...")
            await self.async_refresh_token()

            res = await self._async_get_devices_data()

        return res

    async def _async_get_devices_data(self) -> List[dict]:
        """
        Gets list of device dictonaries from SmartRent's api
        """

        hubs_resp = await self._aiohttp_session.get(
            SMARTRENT_HUBS_URI, headers={"authorization": f"Bearer {self._token}"}
        )
        hubs = await hubs_resp.json()

        if not hubs_resp.ok:
            if hubs.get("errors", [{}])[0].get("code") == "unauthorized":
                raise InvalidAuthError(hubs.get("errors"))

        devices_list = []
        for hub in hubs:
            devices_resp = await self._aiohttp_session.get(
                SMARTRENT_HUBS_ID_URI.format(hub["id"]),
                headers={"authorization": f"Bearer {self._token}"},
            )
            devices = await devices_resp.json()

            for device in devices:
                _LOGGER.info("Found %s: %s", device["id"], device["name"])
                devices_list.append(device)

        return devices_list

    async def async_refresh_token(self) -> None:
        """
        Refreshes API token from SmartRent
        """
        response = {}
        if self._refresh_token:
            response = await self._async_refresh_tokens_via_refresh_token()

            # if refresh token has an error, default to email
            if response.get("errors"):
                codes = [err["code"] for err in response.get("errors")]
                if "unauthorized" in codes:
                    _LOGGER.warning(
                        "Refreshing with refresh_token failed with %s. "
                        "Trying with email and pass instead.",
                        response["errors"],
                    )
                    response = await self._async_refresh_tokens_via_email()
        else:
            response = await self._async_refresh_tokens_via_email()

        if not response.get("errors"):
            self._token = response["access_token"]
            self._refresh_token = response["refresh_token"]
            self._token_exp_time = response["expires"]
            _LOGGER.info("Tokens refreshed!")
        else:
            raise InvalidAuthError(
                "Token not retrieved! "
                f'Loggin probably not successful: {response["errors"]}'
            )

    async def _async_refresh_tokens_via_email(self) -> dict:
        """
        Calls api endpoint to get initial tokens with email and password
        """
        _LOGGER.info("Refreshing tokens with email")
        data = {"email": self._email, "password": self._password}
        resp = await self._aiohttp_session.post(SMARTRENT_SESSIONS_URI, json=data)
        return await resp.json()

    async def _async_refresh_tokens_via_refresh_token(self) -> dict:
        """
        Calls api endpoint to get tokens given a refresh token
        """
        _LOGGER.info("Refreshing tokens with refresh token")
        headers = {"authorization-x-refresh": self._refresh_token}
        resp = await self._aiohttp_session.post(SMARTRENT_TOKENS_URI, headers=headers)
        return await resp.json()

    def _subscribe_device_to_updater(self, device: "Device"):
        """
        Subscribes device to recieve updates
        """
        self._subscribed_devices.add(device)

        if not self._updater_task:
            _LOGGER.info("Starting updater task")
            self._updater_task = asyncio.create_task(self._async_update_state())

        elif self._updater_task:
            if self._updater_task.cancelled():
                _LOGGER.info(
                    "Updater task was previously canceled. Starting updater task again."
                )
                self._updater_task = asyncio.create_task(self._async_update_state())

        if self._ws:
            asyncio.create_task(self._async_ws_joiner(self._ws, device))

    def _unsubscribe_device_to_updater(self, device: "Device"):
        """
        Unsubscribes device to recieve updates
        """
        try:
            self._subscribed_devices.remove(device)
        except KeyError:
            pass

        if self._updater_task:
            if (
                not self._updater_task.cancelled()
                and len(self._subscribed_devices) == 0
            ):
                _LOGGER.info("Device list empty. Stopping updater task for now.")
                self._updater_task.cancel()
                self._ws = None

    async def _async_fetch_subscribed_devices_status(self):
        """
        Calls ``_async_fetch_state`` for all subscribed devices
        """
        _LOGGER.info("Fetching current status for all devices...")
        await asyncio.gather(
            *[device._async_fetch_state() for device in self._subscribed_devices]
        )
        _LOGGER.info("Done fetching data!")

    async def _async_ws_joiner(self, ws, device: "Device"):
        """
        Joins ``Device`` to websocket
        """
        joiner = JOINER_PAYLOAD.format(device_id=device._device_id)
        _LOGGER.info(
            "Joining topic for %s:%s ...",
            device._name,
            device._device_id,
        )
        await ws.send(joiner)

    async def _async_ws_join_devices(self, ws, devices: List["Device"]):
        """
        Takes list of ``Device`` and joins them to websocket
        """
        _LOGGER.info("Joining devices to websocket conn...")
        await asyncio.gather(*[self._async_ws_joiner(ws, device) for device in devices])
        _LOGGER.info("Done Joining devices!")

    async def _async_update_state(self):
        """
        Connects to SmartRent websocket and listens for updates.
        To be ran in the background.

        Calls ``_update`` method for each device when event is found
        """

        retries = 0
        while True:
            try:
                self._ws = None

                _LOGGER.info("Getting new token")
                await self.async_refresh_token()
                token = self._token

                # Update all devices with newest data from regular api
                # we may have missed some stats if websocket was down
                await self._async_fetch_subscribed_devices_status()

                uri = SMARTRENT_WEBSOCKET_URI.format(token)

                _LOGGER.info("Connecting to Websocket...")
                async with websockets.connect(uri) as websocket:

                    # if we connect sucessfully at least one time, reset retries to 0
                    retries = 0
                    self._ws = websocket

                    # Join all devices to websocket connection
                    await self._async_ws_join_devices(
                        websocket, self._subscribed_devices
                    )

                    # iterator to recieve messages from websocket
                    async for message in websocket:
                        message_list = json.loads(f"{message}")
                        formatted_resp = message_list[4]
                        device_id = message_list[2].split(":")[-1]

                        event_type = formatted_resp.get("type", "")
                        event_name = formatted_resp.get("name", "")
                        event_last_read_state = formatted_resp.get(
                            "last_read_state", ""
                        )

                        if event_type:
                            event = (
                                f"{event_type:<15} -> "
                                f"{event_name:<15} -> "
                                f"{event_last_read_state:<20}"
                            )
                            _LOGGER.info(event)

                            for device in self._subscribed_devices:
                                if device._device_id == int(device_id):
                                    await device._update(formatted_resp)
                        else:
                            _LOGGER.info(str(message))

            except Exception as exc:
                _LOGGER.warning(
                    "Exception occured! %s %s", type(exc).__name__, type(exc)
                )
                _LOGGER.warning(traceback.format_exc())

                # set websocket to None
                self._ws = None

                wait_time = 1.25**retries
                wait_time = wait_time if wait_time < 300 else 300
                _LOGGER.warning("Retrying websocket in %s seconds...", wait_time)

                await asyncio.sleep(wait_time)

                retries += 1

    async def _async_send_command(
        self, device: "Device", attribute_name: str, value: str
    ):
        """
        Sends command to SmartRent websocket

        ``attribute_name`` string of attribute to change

        ``value`` value for that attribute to be changed to
        """
        payload = COMMAND_PAYLOAD.format(
            attribute_name=attribute_name, value=value, device_id=device._device_id
        )
        try:
            await self._async_send_payload(device, payload)
        except websockets.exceptions.InvalidStatusCode as exc:
            _LOGGER.debug(
                'Possible issue during send_payload: "%s" '
                "Refreshing token and retrying",
                exc,
            )

            # update token once
            await self.async_refresh_token()

            await self._async_send_payload(device, payload)

    async def _async_send_payload(self, device: "Device", payload: str):
        """
        Sends payload to SmartRent websocket

        ``device`` Device object

        ``payload`` string of device attributes

        Throws ``websockets.exceptions.InvalidStatusCode`` upon bad websocket event
        """
        _LOGGER.info("sending payload %s", payload)

        uri = SMARTRENT_WEBSOCKET_URI.format(self._token)

        async with websockets.connect(uri) as websocket:
            await self._async_ws_joiner(websocket, device)
            await websocket.send(payload)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smartrent']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0', 'websockets>=10.1,<11.0']

setup_kwargs = {
    'name': 'smartrent.py',
    'version': '0.2.4',
    'description': 'Unofficial Python API for SmartRent devices',
    'long_description': "# SmartRent API\n\n[![PyPI version][pypi-version-badge]](https://pypi.org/project/smartrent-py/)\n[![Supported Python versions][supported-versions-badge]](https://pypi.org/project/smartrent-py/)\n[![PyPI downloads monthly][m-downloads-badge]](https://pypistats.org/packages/smartrent-py)\n[![GitHub License][license-badge]](LICENSE.txt)\n[![Documentation Status][docs-badge]](https://smartrentpy.readthedocs.io/en/latest/?badge=latest)\n[![Code style: black][black-badge]](https://github.com/psf/black)\n\nUnofficial api for SmartRent devices\n\nUses websockets for communication and supports async functions\n\nFeel free to ⭐️ this repo to get notified about the latest features!\n\n[📚 Read the docs! 📚](https://smartrentpy.readthedocs.io)\n## Supported Devices\nThis client supports:\n* 🔐 Door Locks\n* 🌡 Thermostats\n* 💧 Leak Sensors\n* 💡 Binary Switches\n\n\n## Usage\n\n### Installation\n\n```bash\npip install smartrent.py\n```\n\n### Getting an API Object\nIn order to get an api object to interact with, you must login with the `async_login` function. This starts and handles a web session with SmartRent.\n\n```python\nimport asyncio\n\nfrom smartrent import async_login\n\nasync def main():\n    api = await async_login('<EMAIL>', '<PASSWORD>')\n\nasyncio.run(main())\n```\n\n### Getting Devices\nYou can get lists of your devices from the api with the `get_locks`, `get_thermostats`, `get_switches` and `get_leak_sensors` functions. You can then interact with the devices with their getter and setter functions.\n\n```python\nimport asyncio\n\nfrom smartrent import async_login\n\nasync def main():\n    api = await async_login('<EMAIL>', '<PASSWORD>')\n\n    lock = api.get_locks()[0]\n    locked = lock.get_locked()\n\n    if not locked:\n        await lock.async_set_locked(True)\n\nasyncio.run(main())\n```\n\n### Automatic Updating\nIf you need to get live updates to your device object from SmartRent, you can do that by calling `start_updater`. You can stop getting updates by calling `stop_updater`.\n\nYou can also set a callback function via `set_update_callback` that will be called when an update is triggered.\n\nFor example, if you want to set your thermostat to `Dad Mode` you can trigger an event every time the `cooling_setpoint` is changed and just change it back to your own desired value.\n```python\nimport asyncio\n\nfrom smartrent import async_login\n\nasync def main():\n    api = await async_login('<EMAIL>', '<PASSWORD>')\n\n    thermo = api.get_thermostats()[0]\n    thermo.start_updater()\n\n    CONSTANT_COOL = 80\n\n    async def on_evt():\n        if CONSTANT_COOL != thermo.get_cooling_setpoint():\n            await thermo.async_set_cooling_setpoint(CONSTANT_COOL)\n\n    thermo.set_update_callback(on_evt)\n\n    while True:\n        await asyncio.sleep(60)\n\nasyncio.run(main())\n```\n\n## Development\n### Setting up dev enviornment\n\n```\npip install -r requirements_test.txt\n```\n\n### Running the code formatter\n[Black](https://github.com/psf/black) is used for quick and easy code formatting\n\n```\nblack smartrent\n```\n\n## Special thanks\nMuch inspiration was taken from these projects:\n\n[AMcPherran/SmartRent-MQTT-Bridge](https://github.com/AMcPherran/SmartRent-MQTT-Bridge)\n\n[natanweinberger/smartrent-python](https://github.com/natanweinberger/smartrent-python)\n\n[Burry/homebridge-smartrent](https://github.com/Burry/homebridge-smartrent)\n## Versioning\n\nWe use [SemVer](http://semver.org/) for versioning. For the versions available, see the [tags on this repository](https://github.com/your/project/tags).\n\n# TODOs\n\n* Add support for Two Factor Auth\n\n[pypi-version-badge]: https://img.shields.io/pypi/v/smartrent-py.svg?logo=pypi&logoColor=FFE873&style=for-the-badge\n[supported-versions-badge]: https://img.shields.io/pypi/pyversions/smartrent-py.svg?logo=python&logoColor=FFE873&style=for-the-badge\n[downloads-badge]: https://static.pepy.tech/personalized-badge/smartrent-py?period=total&units=international_system&left_color=grey&right_color=orange&left_text=total%20downloads&style=for-the-badge\n[m-downloads-badge]: https://img.shields.io/pypi/dm/smartrent-py.svg?style=for-the-badge\n[license-badge]: https://img.shields.io/github/license/ZacheryThomas/smartrent.py.svg?style=for-the-badge\n[docs-badge]: https://readthedocs.org/projects/smartrentpy/badge/?version=latest&style=for-the-badge\n[black-badge]: https://img.shields.io/badge/code%20style-black-000000.svg?style=for-the-badge\n",
    'author': 'Zachery Thomas',
    'author_email': 'zacherythomas12@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zacherythomas/smartrent.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1',
}


setup(**setup_kwargs)



import inspect
import logging

from typing import Union, List, Dict
from .utils import Client

_LOGGER = logging.getLogger(__name__)


class Device:
    """
    Base class for SmartRent devices
    """

    def __init__(self, device_id: Union[str, int], client: Client):
        self._device_id = int(device_id)
        self._name: str = ""
        self._update_callback_funcs = []
        self._updater_task = None

        self._client: Client = client

    def __del__(self):
        self.stop_updater()

    @staticmethod
    def _structure_attrs(attrs: List[Dict[str, any]]) -> Dict[str, any]:
        """
        Converts device json object to hirearchical list of attributes

        ``attrs``: List of device attributes
        """
        structure = {}

        for attr in attrs:
            name = attr.get("name")
            state = attr.get("state")

            structure[name] = state

        _LOGGER.info("constructed attribute structure: %s", structure)
        return structure

    def _fetch_state_helper(self, data: Dict[str, str]):
        """
        Called by ``_async_fetch_state``

        Converts event dict to device param info
        """
        raise NotImplementedError

    async def _async_fetch_state(self):
        """
        Fetches device information from SmartRent api

        Calls ``_fetch_state_helper`` so device can parse out info and update its state.

        Calls function passed into ``set_update_callback`` if it exists.
        """
        _LOGGER.info("%s: Fetching data from device endpoint...", self._name)
        data = await self._client.async_get_devices_data()

        # Find device id that belongs to me then call _fetch_state_helper
        for device in data:
            if device["id"] == self._device_id:
                self._fetch_state_helper(device)

                await self._async_call_callbacks()

    def start_updater(self):
        """
        Allows device to update it's attrs in the background
        """
        self._client._subscribe_device_to_updater(self)

    def stop_updater(self):
        """
        Turns off automatic attr updates
        """
        self._client._unsubscribe_device_to_updater(self)

    def set_update_callback(self, func) -> None:
        """
        Allows callback to be fired when ``_async_update_state``
        or ``_async_fetch_state`` gets new information
        """

        self._update_callback_funcs.append(func)

    def unset_update_callback(self, func) -> None:
        """
        Removes callback from being fired when ``_async_update_state``
        or ``_async_fetch_state`` gets new information
        """

        try:
            self._update_callback_funcs.remove(func)
        except ValueError:
            pass

    def _update_parser(self, event: Dict[str, any]) -> None:
        """
        Called by ``_async_update_state``

        Converts event dict to device attr info
        """
        raise NotImplementedError

    async def _update(self, event: Dict[str, any]):
        """
        Recieves event dict, calls ``_update_parser`` for each device and callbacks
        """

        # handle updating of device attrs
        self._update_parser(event)

        # handle calling callbacks
        await self._async_call_callbacks()

    async def _async_call_callbacks(self):
        """
        Handles calling all callbacks
        """
        for func in self._update_callback_funcs:
            if inspect.iscoroutinefunction(func):
                await func()
            else:
                func()
